import os
import json
import requests
from datetime import datetime, timezone
import copy
import csv

class utils:

         
    @staticmethod
    def collect_token(base_url: str, c_key: str, c_secret: str, token_store_path: str = '/tmp/sf_token_store.json', key_tag: str = 'default') -> str: 
        """
        #### Inputs: 
            -@base_url: Salesforce org url (i.e 'https://examplecompany.my.salesforce.com')
            -@c_key: consumer key of connected app
            -@c_secret: consumer secret of connected app 
            -@token_store_path: optional, location to store resulting token
        #### Expected Behaviour:
            - Is a wrapper for check_token_store and get_access_token usage. 
            - Checks if there is a cached token locally, otherwise retrieves a new one and stashes it before returning it. 
        #### Returns: 
            - An access token 
        #### Side Effects: 
            - stores a json file in the token_store_path
        #### Exceptions: 
            - None
        """
        check_token = utils.check_token_store(key_tag=key_tag, token_store_path=token_store_path)
        if check_token == None:
            access_token = utils.get_access_token(base_url, c_key, c_secret, token_store_path)
        else:
            access_token = check_token
        return(access_token)


    @staticmethod
    def check_token_store(key_tag: str = 'default', token_store_path: str = '/tmp/sf_token_store.json',) -> str | None:  
        """
        #### Inputs: 
            -@key_tag: the tag to look under in the token store
            -@token_store_path: optional, location to check for token 
        #### Expected Behaviour: 
            - look for a token_store file in the token_store_path under the key_tag, 
            if a token is found it will be returned,
            otherwise return None
        #### Returns: 
            - str: a valid access token
            - None: if no valid access token is found 
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None 
        """
        if os.path.isfile(token_store_path):
            with open(token_store_path, 'r') as json_file:
                loaded_json = json.load(json_file)
                test_access_token = loaded_json.get(key_tag, {}).get("access_token", None)
                return(test_access_token)
        return None


    @staticmethod
    def get_access_token(base_url: str, c_key: str, c_secret: str, token_store_path: str = '/tmp/sf_token_store.json', key_tag: str = 'default') -> str:
        """
        #### Inputs:
            -@base_url: Salesforce org url (i.e 'https://examplecompany.my.salesforce.com') 
            -@c_key: consumer key of connected app
            -@c_secret: consumer secret of connected app
            -@token_store_path: optional, location to store token
            -@key_tag: the tag to store the key under in the token store
        #### Expected Behaviour: 
            - request an access token using the consumer key and secret. If non-200 response code
            raise an exception, otherwise save the token to the token_store_path 
            under the path [key_path]['access_token'], alongside a timestamp, and return the access token.
        #### Returns: 
            - An access token
        #### Side Effects: 
            - Save json data to token_store_path
        #### Exceptions: 
            - 'Token Request Error: ...' If the token retrieval request returns non-200, exception raised with the content. 
        """
        urlstring = f"{base_url}/services/oauth2/token?grant_type=client_credentials&client_id={c_key}&client_secret={c_secret}"
        resp = requests.post(urlstring)
        if resp.status_code != 200:
            raise Exception(f'Token Request Error: {str(resp.content)}')
        content = json.loads(resp.content)
        access_token = content["access_token"]
        with open(token_store_path, "w") as token_file:
            token_json = {key_tag: {"access_token": access_token, 'timestamp': str(datetime.now(timezone.utc))}}
            token_file.write(json.dumps(token_json))
        return(access_token)
    

    @staticmethod
    def flush_token_store(token_store_path: str = '/tmp/sf_token_store.json'):
        """
        #### Inputs: 
            -@token_store_path: optional, location of token store
        #### Expected Behaviour: 
            - Remove the file that the token is stored in if it exists
        #### Returns: 
            - None 
        #### Side Effects: 
            - Removes token_store_path file if present
        #### Exceptions: 
            - None 
        """
        if os.path.isfile(token_store_path):
            os.remove(token_store_path)


    @staticmethod
    def delete_single_access_token(key_tag: str = 'default', token_store_path: str = '/tmp/sf_token_store.json'):
        """
        #### Inputs: 
            -@key_tag: the tag associated with the access token you want to delete
            -@token_store_path: optional, location tokens are stored
        #### Expected Behaviour: 
            - if the token_store_path is found and the key_tag is in the json, delete the 
            key and save back to the token store
        #### Returns: 
            - None 
        #### Side Effects: 
            - Modifies the file stored at token_store_path 
        #### Exceptions: 
            - None 
        """
        json_data = {}
        if os.path.isfile(token_store_path): ## can't read and write at the same time
            with open(token_store_path, 'r') as store_file:
                json_data = json.load(store_file)
            json_data.pop(key_tag)
            with open(token_store_path, 'w') as w:
                json.dump(json_data, w)


    @staticmethod
    def retrieve_full_token_store(token_store_path: str = '/tmp/sf_token_store.json') -> dict:
        """
        #### Inputs: 
            -@token_store_path: optional, location tokens are stored
        #### Expected Behaviour: 
            - If the token_store_path is found then return the dict of values in it,
            if it's not found then raise an exception. 
            Useful for sanity check that you're using the right token store
        #### Returns: 
            - the dict stored in the token store
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None 
        """
        if not os.path.isfile(token_store_path):
            raise Exception('No file found in token_store_path')
        return_dict = {}
        with open(token_store_path, 'r') as json_file: 
            return_dict = json.load(json_file)
        return(return_dict)
    
    
    @staticmethod
    def get_z_time(year: int, month: int, day:int) -> str:
        """
        #### Input: 
            -@year: year: yyyy
            -@month: month: mm
            -@day: day: dd
        #### Expected Behaviour:
            - For use in filtering salesforce using a datetime, which requires a zero offset or 'z' time 
            format. Passing in the year, month, day it will convert to this zero offset format 
            and return the string
        #### Returns: 
            - str: the zero offset time string
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None 
        """
        unformatted = datetime(year, month, day, 0 ,0, tzinfo=timezone.utc)
        formatted = str(unformatted).replace('+00:00', 'Z')
        formatted = formatted.replace(' ', 'T')
        return(formatted)
    
    @staticmethod   
    def split_nested_record_list(input_list: list[dict]) -> dict[str,list[dict]]: #tesed
        """
        #### Input:    
            - a list of dicts resulting from a nested query to sf data 
        #### Expected Behaviour: 
            - for each record in the list; 
            each record is appended to a list stored in the return dict, under the key of its type
            each record will have its parent record in it as an additional field, they key being the type of the parent
            and the value being its ID 
        #### Returns:
            - a dict{str: list[dict]} key being the type of the record, value being a list of records
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None  (this isn't true)
        """
        output_dict = {}

        def split_record(record: dict, parent_id: str, parent_type: str):
            try:
                record_type = record['attributes']['type']
                record_id = record['Id']
            except:
                raise Exception('record Type or Id not found')
            current_record = {}
            if parent_id:
                current_record[parent_type] = parent_id
            for key, value in record.items():
                if key == 'attributes' or value == None:
                    continue
                elif type(value) == dict:
                    if value.get('attributes') != None: #
                        split_record(record=value, parent_id=record_id,parent_type=record_type)
                    for nested_record in value.get('records', []):
                        split_record(record=nested_record, parent_id=record_id, parent_type=record_type)
                else:
                    current_record[key] = value
            output_dict[record_type] = output_dict.get(record_type, []) + [current_record]

        for record in input_list:
            split_record(record,None,None)
        return(output_dict)
    
    
    @staticmethod
    def build_in_querystring(querystring: str, in_list: list[str]) -> tuple[str, list[str]]: 
        """
        #### Inputs: 
            -@querystring: soql query containing <in> substring 
            -@in_list: list of strings to place within the (in) part of the query
        #### Expected Behaviour: 
            - Builds the longest in query it can from the in_list, then returns the resulting 
                querystring and the remaining in_list elements  
        #### Returns: 
            -str: the built querystring 
            -list[str]: the remaining elements in the list after the maximum size query string was constructed
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None 
        """ 
        index = 0 
        while index + 1 <= len(in_list) and len(querystring.replace('<in>', f'({",".join(in_list[:index+1])})')) < 20000:
           index += 1
        return(querystring.replace('<in>', f'({",".join(in_list[:index])})'), in_list[index:])
    

    @staticmethod
    def combine_nested_result_dicts(source_dict: dict, destination_dict: dict) -> dict:
        """
        #### Inputs:
            -@source_dict: the dict to copy FROM
            -@destination_dict: the dict to copy TO
        #### Expected Behaviour: 
            - For all keys in the dict, append values from the source dict to the values in the destination dict
            - if the source dict has a key not present in the destination dict, it will create one
        #### Returns: 
            - the combined dict 
        #### Side Effects:  
            - None 
        #### Exceptions: 
            - None 
        """
        s_dict = copy.deepcopy(source_dict)
        d_dict = copy.deepcopy(destination_dict)
        for key, value in s_dict.items():
            destination_value = d_dict.get(key, [])
            destination_value.extend(value)
            d_dict[key] = destination_value
        return(d_dict)
    

    @staticmethod 
    def build_key_list(dict_list: list[dict]) -> list:
        """
        #### Inputs: 
            -@dict_list: a list of dicts
        #### Expected Behaviour: 
            - Loop through all dicts in the list, getting the unique keys in them
                all and then return the list
        #### Returns: 
            - Unique list of keys 
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None 
        """
        header_list = [] # set is not used because we want to preserve order
        for record in dict_list:
            for key in record.keys():
                if key not in header_list:
                    header_list.append(key)
        return(header_list)
    
    @staticmethod
    def record_list_to_csv(record_list: list[dict], output_filename: str, append: bool = False):
        """
        #### Inputs: 
            -@record_list: a list of dicts, representative of a non-nested query result
            -@output_filename: the filename to save the resulting .csv as (don't include .csv)
            -@append: if True, will append to an existing file, else writes 
        #### Expected Behaviour: 
            - if append=True, open the existing file and use the combination of the existing fieldnames 
                and any potential new fieldnames from record_list 
            - otherwise, build the fieldnames using build_key_list
            - write/append to the csv file
        #### Returns: 
            - None 
        #### Side Effects: 
            - Writes/appends to the output_filename
        #### Exceptions: 
            - None 
        """
        fieldnames = []
        records = []
        if append:
            with open(f'{output_filename}.csv', 'r') as existing_file: 
                reader = csv.DictReader(existing_file)
                for row in reader:
                    records.append(row)

        records.extend(record_list)
        print(records)
        fieldnames = utils.build_key_list(records)
        print(f'fieldnames = {fieldnames}')
        with open(f'{output_filename}.csv', 'w') as f:
            w = csv.DictWriter(f, fieldnames)
            w.writeheader()
            w.writerows(records)

    @staticmethod
    def record_list_dict_to_csv(record_list_dict: dict, filename_prefix: str,  append: bool = False):
        """
        #### Inputs: 
            -@record_list_dict: dict of lists of dicts, keys are used in the filename, 
            lists of dicts are saved as CSVs, each dict being a row,
            -@filename_prefix: a a prefix to prepend to each filename, which will be followed by its key
            -@append:  if True, will append to an existing file, else writes 
        #### Expected Behaviour: 
            - Each key in the dict is used to create a new csv and to call records_to_csv, 
                passing in a filename created from a combination of the filename_prefi and the corresponding 
                key in the dict used to access that dict list 
        #### Returns:
            - None 
        #### Side Effects: 
            - Creates csv files for each key in the dict 
        #### Exceptions: 
            - None 
        """
        for key, value in record_list_dict.items():
            filename = f'{filename_prefix}_{key}.csv'
            utils.record_list_to_csv(record_list=value, output_filename=filename, append=append)   


    @staticmethod
    def combine_records(record_one: dict, record_two: dict) -> dict: 
        """
        #### Inputs: 
            -@record_one: a dict
            -@record_two: a dict
        #### Expected Behaviour: 
            - Provided two dictionaries will add keys and values from record_two to record_one,
            - it only bothers about keeping record_one items in the case of a collision of keys
        #### Returns: 
            - the combined dict 
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None 
        """
        return_record = {}
        return_record.update(record_one)
        for key in [x for x in record_two if x not in return_record.keys()]:
            return_record[key] = record_two[key]
        return(return_record)



from unittest.mock import Mock, patch, call
import requests
import json
from unittest import TestCase
import os
import csv
from sftocsv.utils import utils
import shutil
class test_utils(TestCase):
    
    ### check_token_store tests 
    def test_check_token_store(self):
        """
        #### Function: 
            - utils.check_token_store
        #### Inputs: 
            3 Calls: 
            1. -@key_tag: not entered ('default'): this is in the token store
            2. -@key_tag: 'missing': this is not in the token store
            --- then the token store is deleted
            3. -@key_tag: not entered ('default')
        #### Expected Behaviour: 
            -1. the access token is found in the token store and returned
            -2. the key tag isn't found so the access token isn't found, so None is returned
            -3. the token store isn't found, so None is returned
        #### Assertions: 
            - first call returns 'test_token'
            - second call returns None 
            - third call returns None 
        """
        ### setup
        test_token_store = {'default': {'access_token': 'test_token'}}
        with open('/tmp/sf_token_store.json', 'w') as token_file:
            json.dump(test_token_store, token_file)
        ### 
        assert(utils.check_token_store() == 'test_token')
        assert(utils.check_token_store(key_tag='missing') == None)
        os.remove('/tmp/sf_token_store.json')
        assert(utils.check_token_store() == None)
        

    ### --- get_access_token tests 
    @patch("requests.post")
    def test_get_access_token(self, mock_post):
        """
        #### Function: 
            - utils.get_access_token
        #### Inputs: 
            -@base_url: 'https://localhost/ (not important as post is mocked)
            -@c_key: test string 
            -@c_secret: test string
            -@token_store_path: not passed in
            -@key_tag: not passed in
        #### Expected Behaviour:
            - Callout is made, returns 200 and an access_token in the response body
            - access_token is saved to the token_store_path under the key_tag ('default')
            - access token is returned
        #### Assertions: 
            - access token returned from the request is stored in the token_store_path
            - value returned is the access token
        """
        ### Setup
        os.mkdir('testing_folder')
        os.chdir('testing_folder')
        ### 
        mock_response = requests.Response()
        mock_response.status_code = 200
        mock_response._content = b"""{"access_token":"fake_token",
                                       "signature":"fake_signature",
                                       "scope":"chatter_api api",
                                       "instance_url":"fake_instance_url",
                                       "id":"fake_id",
                                       "token_type":"Bearer","issued_at":"1687833980742"}"""

        mock_post.return_value = mock_response
        resp = utils.get_access_token(base_url='https://localhost/', c_key='test_key', c_secret='test_secret')
        with open('/tmp/sf_token_store.json', 'r') as token_file:
            loaded_json = json.load(token_file)
            file_access_token = loaded_json['default']["access_token"]
            assert file_access_token == "fake_token"
        assert(resp == 'fake_token')
        os.chdir('..')
        shutil.rmtree('testing_folder')

    @patch("requests.post")
    def test_get_access_token_error(self, mock_post):
        """
        #### Function: 
            - utils.get_access_token
        #### Inputs: 
            -@base_url: 'https://localhost/' (not important as post is mocked)
            -@c_key: test string 
            -@c_secret: test string 
            -@token_store_path: not passed in
            -@key_tag: not passed in
        #### Expected Behaviour: 
            - callout is made, the mocked response code is 400 so the exception is thrown
        #### Assertions: 
            - the exception is raised with the expected content
        """
        mock_response = requests.Response()
        mock_response.status_code = 400
        json_string = {"error": "invalid_client", "error_description": "invalid client credentials"}
        mock_response._content = json.dumps(json_string)
        mock_post.return_value = mock_response
        with self.assertRaises(Exception) as context:
            utils.get_access_token(base_url='https://localhost/', c_key='test key', c_secret='test_secret')
        assert str(context.exception) == f'Token Request Error: {json.dumps(json_string)}'

    
    ### --- collect_token tests
    @patch.object(utils, 'get_access_token')
    @patch.object(utils, 'check_token_store')
    def test_collect_token_none_check_token(self, mock_check_token_store, mock_get_access_token):
        """
        #### Function: 
            - utils.collect_token
        #### Inputs:
            -@base_url: 'https://localhost/' (not important as functions using it are mocked)
            -@c_key: test key
            -@c_secret: test_secret
            -@token_store_path: not passed in
        #### Expected Behaviour:
            - utils.check_token_store is mocked to return None, so the following condition is true
            utils.get_access_token returns a value, which becomes access_token, which is returned
        #### Assertions: 
            - the returned value is the string returned by utils.get_access_token
        """
        mock_check_token_store.return_value = None
        mock_get_access_token.return_value = 'access_token'
        assert(utils.collect_token(base_url='https://localhost/', c_key='test key', c_secret='test secret') == 'access_token')

    @patch.object(utils, 'check_token_store')
    def test_collect_token_check_token_returned(self, mock_check_token_store):
        """
        #### Function:
            - utils.collect_token
        #### Inputs: 
            -@base_url: 'https://localhost/ (not important as functions using it are mocked)
            -@c_key: test key
            -@c_secret: test_secret
            -@token_store_path: not passed_in
        #### Expected Behaviour: 
            - check_token_store is mocked to return a non-None value, so it is passed to access_token,
                which is then returned
        #### Assertions: 
            - The returned value is the mocked access token response
        """
        mock_check_token_store.return_value = 'test_token'
        assert(utils.collect_token(base_url='https://localhost/', c_key='test key', c_secret='test secret') == 'test_token')


    ### --- flush_token_store tests --- 
    def test_flush_token_store(self):
        """
        #### Function: 
            - utils.flush_token_store
        #### Inputs: 
            - none passed in on first call
            - 'test_path.json' passed in on second call
        #### Expected Behaviour: 
            - 2 calls: 
            1. No file is found in the token path, nothing is done 
            2. file is found in token path, it's removed
        #### Assertions: 
            - The directory is unchanged after the first call,
            - the token store is deleted after the second call
        """
        ### Setup
        os.mkdir('testing_folder')
        os.chdir('testing_folder')
        with open('test_path.json', 'w') as w:
            w.write('t')
        ###
        utils.flush_token_store()
        assert(os.listdir() == ['test_path.json'])
        utils.flush_token_store('test_path.json')
        assert(os.listdir() == [])
        os.chdir('..')
        shutil.rmtree('testing_folder')

    
    ### --- delete_single_access_token tests --- 
    def test_delete_single_access_token(self):
        """
        #### Function:
            - utils.delete_single_access_token
        #### Inputs:
            -@key_tag: 'default'
            -@token_store_path: 'test_store.json'
        #### Expected Behaviour:
            - the test_store.json has multiple keys stored in it, 
                the file is opened and only the 'default' key is removed
        #### Assertions: 
            - the file is the same as before the function call, minus the 'default' key. 
        """
        ### Setup
        os.mkdir('testing_folder')
        os.chdir('testing_folder')
        example_store = {
            'default': {'access_token': 'default_key', 'timestamp': 'test'},
            'other_key': {'access_token': 'other_key', 'timestamp': 'test'}
        }
        with open('test_store.json', 'w') as w:
            json.dump(example_store, w)
        ### 
        utils.delete_single_access_token(key_tag='default', token_store_path='test_store.json')
        with open('test_store.json', 'r') as r:
            data = json.load(r)
            assert(data == {'other_key': {'access_token': 'other_key', 'timestamp': 'test'}})
        os.chdir('..')
        shutil.rmtree('testing_folder')

    
    ### --- retrieve_full_token_store --- 
    def test_retrieve_full_token_store_exception(self):
        """
        #### Function: 
            - utils.retrieve_full_token_store
        #### Inputs: 
            -@token_store_path: 'test_store.json' (no file is present here)
        #### Expected Behaviour: 
            - Because the token path file is not found, an exception is thrown
        #### Assertions: 
            - The expected exception is thrown
        """
        ### Setup
        os.mkdir('testing_folder')
        os.chdir('testing_folder')
        ### 
        with self.assertRaises(Exception) as context:
            utils.retrieve_full_token_store('test_store.json')
        assert(str(context.exception) == 'No file found in token_store_path')
        os.chdir('..')
        shutil.rmtree('testing_folder')

    def test_retrieve_full_token_store(self):
        """
        #### Function: 
            - utils.retrieve_full_token_store
        #### Inputs: 
            -@token_store_path: 'test_store.json' (file is present here)
        #### Expected Behaviour: 
            - The file is found and opened, then the loaded json is returned
        #### Assertions: 
            - the contents of test_store.json are returned.
        """
        ### Setup
        os.mkdir('testing_folder')
        os.chdir('testing_folder')
        example_store = {
            'default': {'access_token': 'default_key', 'timestamp': 'test'},
            'other_key': {'access_token': 'other_key', 'timestamp': 'test'}
        }
        with open('test_store.json', 'w') as w:
            json.dump(example_store, w)
        ### 
        resp = utils.retrieve_full_token_store('test_store.json')
        assert(resp == example_store)
        ### Cleanup
        os.chdir('..')
        shutil.rmtree('testing_folder')
        ###


    ### --- get_z_time tests ---
    def test_get_z_time(self): 
        """
        #### Function:
            - utils.get_z_time
        #### Inputs:
            -@year: 2005
            -@month: 10
            -@day: 08
        #### Expected Behaviour: 
            - the input year month and date will be converted into a datettime, which will then be formated to zero-offset time
            and returned
        #### Assertions: 
            - the returned string is as expected
        """
        returned_time = utils.get_z_time(2005, 10, 8)
        assert(returned_time == '2005-10-08T00:00:00Z')

    
    ### --- split_nested_record tests ---
    def test_split_nested_record_flat(self):
        """
        #### Function: 
            - utils.split_nested_record_list
        #### Input/s:
            - a list of dicts that have no further nesting, they are also all of the same type 'example'
        #### Expected Behaviour: 
            - For each record split_record will be called, none of the values are dicts (other than attributes,
            which is skipped),
            so the items are copied into one list and this is stored in the returned dict under the key 'example'
        #### Assertions:
            - the returned dict has the one key 'example' which stores the expected list of dicts
        """
        input_list = [
            {'Id': 'id1',
             'attributes': {'type': 'example'},
             'key1': 'valuea1',
             'key2': 'valuea2'
             },
            {'Id': 'id2',
             'attributes': {'type': 'example'},
             'key1': 'valueb1',
             'key2': 'valueb2'
            },
            {'Id': 'id3',
             'attributes': {'type': 'example'},
             'key1': 'valuec1',
             'key2': 'valuec2'
            }
        ]
        expected_result = [
            {'Id': 'id1',
             'key1': 'valuea1',
             'key2': 'valuea2'
             },
            {'Id': 'id2',
             'key1': 'valueb1',
             'key2': 'valueb2'
            },
            {'Id': 'id3',
             'key1': 'valuec1',
             'key2': 'valuec2'
            }
        ]
        resp_dict = utils.split_nested_record_list(input_list)
        assert(len(resp_dict) == 1)
        assert(resp_dict['example'] == expected_result)
    
    def test_split_nested_record_no_type_no_id(self):
        """
        #### Function: 
            - utils.split_nested_record_list
        #### Input/s:
            - first call it with a list, the only element is a dict with an Id but NO type key
            in the attributes,
            - second call with a list, the only element is a dict with a type key in attributes, 
            but no Id key anywhere
        #### Expected Behaviour:
            - Both calls should be caught in the first and second lines respectively in the try-except of 
            split_record function
        #### Assertions:
            - Both calls result in the same exception
        """
        type_missing_input = [
            {'Id': 'id1',
             'attributes': {'none': 'none'},
             'key1': 'value1',
             'key2': 'value2'}
        ]
        with self.assertRaises(Exception) as context:
            utils.split_nested_record_list(type_missing_input)
        assert(str(context.exception) == 'record Type or Id not found')
        id_missing_input = [
            {'key1': 'value1',
             'attributes': {'type': 'example'},
             'key2': 'value2'}
        ]
        with self.assertRaises(Exception) as context:
            utils.split_nested_record_list(id_missing_input)
        assert(str(context.exception)) == 'record Type or Id not found'

    def test_split_nested_record_2d(self):
        """
        #### Function: 
            - utils.split_nested_record_list
        #### Input/s:
            - a list of dicts that have a dict as a value in them,
                this dict has a 'records' key and the value is a list of dicts, 
        #### Expected Behaviour:
            - Each of the dicts found in 'records' is added to their corresponding type key 
            and returned, as are the outer. 
            The id1 and id4 records are added under the 'outer1' key, 
            the key 'inners' has a dict value and the 'records' key is within, 
            3 of these records are of type 'inner1' and are added to this key in the return dict
            the key 'inners2' has a records list as well, both records are type 'inner2' and are added to this key. 
        #### Assertions: 
            - The returned dict has 2 records in the list stored in 'outer1', they should only have Id and key1 items inside
            - it should have 3 records in the list stored in 'inner1', they should have id and key1 items 
            - it should have 2 records in the list stored in 'inner2, with id and key1 items
            - inner1 and inner2 records will have a key in them 'outer1' with the id of their parent in this value 
        """
        input_list = [
            {'Id': 'id1',
            'key1': 'valueA',
             'attributes':{'type': 'outer1'},
             'inners': {'records': [{'Id': 'id2',
                          'key1': 'valueB',
                          'attributes': {'type': 'inner1'}},
                          {'Id': 'id3',
                           'key1': 'valueC',
                           'attributes': {'type': 'inner1'}}]}
            },
            {'Id': 'id4',
            'key1': 'valueD',
             'attributes':{'type': 'outer1'},
             'inners': {'records': [{'Id': 'id5',
                          'key1': 'valueE',
                          'attributes': {'type': 'inner1'}}]},
            'inners2': {'records': [{'Id': 'id6',
                          'key1': 'valueF',
                          'attributes': {'type': 'inner2'}},
                          {'Id': 'id7',
                           'key1': 'valueG',
                           'attributes': {'type': 'inner2'}}]}
            }]
        resp = utils.split_nested_record_list(input_list)
        assert(len(resp) == 3)
        assert(resp['outer1'] == [{'Id': 'id1', 'key1': 'valueA'},
                                {'Id': 'id4', 'key1': 'valueD'}])
        assert(resp['inner1'] == [{'outer1': 'id1', 'Id': 'id2', 'key1': 'valueB'},
                                  {'outer1': 'id1','Id': 'id3', 'key1': 'valueC'},
                                  {'outer1': 'id4','Id': 'id5', 'key1': 'valueE'}])
        assert(resp['inner2'] == [{'outer1': 'id4', 'Id': 'id6', 'key1': 'valueF'},
                                  {'outer1': 'id4', 'Id': 'id7', 'key1': 'valueG'}])
        
    def test_split_nested_record_3d(self):
        """
        #### Function:
            - utils.split_nested_record_list
        #### Input/s:
            - a list of dicts, the one dict inside has a nested list of records, a record in the list 
            has a nested list of records itself .
        #### Expected Behaviour: 
            - the outer dict is saved into 'outer' key, 
            - the records nested 1 deep are saved into 'middle' key, it has the outer key saved in 'outer'
            - the records nested 2 deep are saved into 'inner' key, they have the middle key saved in 'middle'
        #### Assertions:
            - The returned dict has the keys 'outer', 'middle', 'inner'
            - outer has 1, middle 2, inner 2
            - outer, middle, inner all match expected values
        """
        input_list = [{'Id': 'id1',
                       'attributes': {'type': 'outer'},
                       'key1': 'valueA',
                       'inners': {'records': [
                            {'Id': 'id2',
                             'attributes': {'type': 'middle'},
                             'key1': 'valueB',
                             'inners': {'records': [
                                 {'Id': 'id4',
                                  'attributes': {'type': 'inner'},
                                  'key1': 'valueD'},
                                 {'Id': 'id5',
                                  'attributes': {'type': 'inner'},
                                  'key1': 'valueE'}
                             ]}},
                            {'Id': 'id3',
                             'attributes': {'type': 'middle'},
                             'key1': 'valueC'}]
                                }
                        }]
        resp = utils.split_nested_record_list(input_list)
        assert(len(resp) == 3)
        assert(resp['outer'] == [{'Id': 'id1', 'key1': 'valueA'}])
        assert(resp['middle'] == [{'outer' : 'id1', 'Id': 'id2', 'key1': 'valueB'},
                                  {'outer' : 'id1', 'Id': 'id3', 'key1': 'valueC'}])
        assert(resp['inner'] == [{'middle': 'id2', 'Id': 'id4', 'key1': 'valueD'},
                                 {'middle': 'id2', 'Id': 'id5', 'key1': 'valueE'}])
        
    def test_split_nested_record_list_linked(self):
        """
        #### Function:
            - utils.split_nested_record_list
        #### Input:
            -@input_list: a dict with a nested list of records under 'linkage', 
                there is 1 linkage record and it has a 'linked' record as a single child, not in a 
                records list. An example of how a record like LoanContactRole would be returned, a
                one-to-one relation of nesting
        #### Expected Behaviour:
            - the outer dict is saved to the 'outer' key, 
            - the record found in the 'records' key of 'linkage' type is added to the 'linkage' key,
            - the record found in the 'inner__r' key of the linkage record is of type 'inner' and is added to this key
        #### Assertions:
            - returned dict has 1 record in 'outer' key list
            - 1 record in 'linkage' , it has 'id1' in the 'outer' key (its parent's id)
            - 1 record in 'inner', it has 'id2' in the 'linkage' key (its parent's id)
        """
        input_list = [{'Id': 'id1', 
                        'attributes': {'type': 'outer'},
                        'key1': 'valueA',
                        'linkage': {'records': [
                            {'Id': 'id2',
                            'attributes': {'type': 'linkage'},
                            'key1': 'valueB',
                            'inner__r': {
                                'Id': 'id3',
                                'attributes': {'type': 'inner'},
                                'key1': 'valueC'
                            }}
                        ]}
                    }]
        resp = utils.split_nested_record_list(input_list)
        assert(len(resp) == 3)
        assert(resp['outer'] == [{'Id': 'id1', 'key1': 'valueA'}])
        assert(resp['linkage'] == [{'Id': 'id2', 'key1': 'valueB', 'outer': 'id1'}])
        assert(resp['inner'] == [{'Id': 'id3', 'key1': 'valueC', 'linkage': 'id2'}])

    
    ### --- build_in_querystring tests ---
    def test_build_in_querystring_smaller(self):
        """
        #### Function: 
            - utils.build_in_querystring
        #### Inputs: 
            -@querystring: a string with <in> present in it 
            -@in_list: a list of 100 elements, which total a length of < 1000 (and therefore less than 20,000)
        #### Expected Behaviour: 
            - The while loop increments and the second condition is never true, so it goes until the 
                the first increment is false (reaches the end of the list), then returns the querystring and an empty remaining list
        #### Assertions: 
            - the returned string contains all of the elements of the in_list
            - the returned list of remaning elements is empty 
            - Assert the section of the query before the (in) part matches the input of that part
        """
        ### Setup
        input_list = [str(x) for x in range(50,151)]
        ###
        returned_query, remaining_elements = utils.build_in_querystring(querystring='select name from numbers__c where id in <in>', in_list = input_list)
        in_section = returned_query.partition('(')[2]
        in_section = in_section.partition(')')[0]
        assert(in_section.split(',') == input_list)
        assert(remaining_elements == [])
        assert('select name from numbers__c where id in ' == returned_query.partition('(')[0])
    
    def test_build_in_querystring_bigger(self):
        """
        #### Function: 
            - uitls.build_in_querystring
        #### Inputs: 
            -@querstring: a string with <in> present in it
            -@in_list: a list of 10000 elements with a total length > 20,000 (they're all >=3 digits)
        #### Expected Behaviour: 
            - The while loop increments until the second condition is true, (which it will be at some point due to the length of the in_list),
                then the resulting string is returned, and the remaining unused elements are returned in a list as well 
        #### Assertions: 
            - The returned string could not accomodate 1 more value (+ a comma), i.e it's  >= 19996 characters long but less than 20000  
            - the combination of elements inside the query and returned sum to equal the in_list
        """
        ### Setup
        input_list = [str(x) for x in range(100, 10101)]
        ###
        returned_query, remaining_elements = utils.build_in_querystring(querystring='select name from numbers__c where id in <in>', in_list=input_list)
        in_section = returned_query.partition('(')[2]
        in_section = in_section.partition(')')[0]
        assert(in_section.split(',') + remaining_elements == input_list)
        assert(20000 > len(returned_query) >= 19996)


    ### --- combine_nested_result_dicts tests ---
    def test_combine_nested_result_dicts(self):
        """
        #### Function: 
            - utils.combine_nested_result_dicts
        #### Inputs: 
            -@source_dict: a non-empty dict 
            -@destination_dict: a non-empty dict with some overlap of keys 
        #### Expected Behaviour: 
            - the keys that are common between the dicts will result in a combined list of dicts in the values,
                 the keys only in the source dict will be added as new keys and value in the destination
        #### Assertions: 
            - the output dict matches expected value 
        """
        source = {'key1': [{'record1': '1', 'name': 'test'}, {'record2': '2', 'name': 'test2'}],
                  'key2': [{'record3': '3', 'name': 'test3'}]}
        destination = {'key2': [{'record4': '4', 'name': 'test4'}],
                       'key3': [{'record5': '5', 'name': 'test5'}, {'record6': '6', 'name': 'test6'}]}
        resp = utils.combine_nested_result_dicts(source_dict=source, destination_dict=destination)
        assert resp == {'key1': [{'record1': '1', 'name': 'test'}, {'record2': '2', 'name': 'test2'}],
                  'key2': [{'record4': '4', 'name': 'test4'}, {'record3': '3', 'name': 'test3'}],
                  'key3': [{'record5': '5', 'name': 'test5'}, {'record6': '6', 'name': 'test6'}]}
        

    ### --- build_key_list tests --- 
    def test_build_key_list(self):
        """
        #### Function: 
            - utils.build_key_list
        #### Inputs: 
            -@dict_list: a list of dicts with some overlapping and some shared keys
        #### Expected Behaviour: 
            - the list is looped through and each unique key is added to the returned list 
        #### Assertions: 
            - the returned list has a list of 1 instance of each key in the dict list 
        """
        input_list = [{'key1': 'value1', 'key2': 'value2', 'key3': 'value3'},
                      {'key3': 'value3', 'key4': 'value4', 'key5': 'value5'},
                      {'key1': 'value1', 'key4':'value4', 'key6':'value6'}]
        assert(utils.build_key_list(input_list) == ['key1', 'key2', 'key3', 'key4', 'key5', 'key6'])

    
    ### --- record_list_to_csv tests --- 
    def test_record_list_to_csv_write(self):
        """
        #### Function:
            - utils.record_list_to_csv
        #### Inputs: 
            -@record_list: a list of records
            -@output_filename: 'test_file.csv' 
            -@append: false
        #### Expected Behaviour: 
            - Because append is false, only the new record_list is used to build the fieldnames,
            - each dict in the list is written as a row to the file using csv.DictWriter
        #### Assertions: 
            - The output_filename when opened and read back into a dict list contains all of the values of record_list
        """
        ### Setup
        os.mkdir('testing_folder')
        os.chdir('testing_folder')
        ### 
        input_list = [{'key1': 'value11', 'key2': 'value21', 'key3': 'value31'},
                      {'key1': 'value12', 'key2': 'value22', 'key3': 'value32'},
                      {'key1': 'value13', 'key2': 'value23', 'key3': 'value33', 'key4': 'value34'}]
        utils.record_list_to_csv(record_list=input_list, output_filename='test_file')
        expected_list = [{'key1': 'value11', 'key2': 'value21', 'key3': 'value31', 'key4': ''},
                      {'key1': 'value12', 'key2': 'value22', 'key3': 'value32', 'key4': ''},
                      {'key1': 'value13', 'key2': 'value23', 'key3': 'value33', 'key4': 'value34'}]
        output_list = []
        with open('test_file.csv', 'r') as r:
            reader = csv.DictReader(r)
            for row in reader: 
                output_list.append(row)
        assert(output_list == expected_list)
        os.chdir('..')
        shutil.rmtree('testing_folder')

    def test_record_list_to_csv_append(self):
        """
        #### Function:
            - utils.record_list_to_csv
        #### Inputs: 
            -@record_list: a list of records 
            -@output_filename: 'test_file.csv'
            -@append: True
        #### Expected Behaviour: 
            - because append is true, the output_filename is read first to build the fieldnames, 
                then the incoming record_list is used with build_key_list and combined with fieldnames 
            - each dict in the list is added as a row to the file 
        #### Assertions: 
            - the output_filename when opened and read back into a dict list contains all the expected values
        """
        ### Setup
        os.mkdir('testing_folder')
        os.chdir('testing_folder')
        prior_list = [{'key1': 'value11', 'key2': 'value12', 'key3': 'value13'},
                          {'key1': 'value21', 'key2': 'value22', 'key3': 'value23'}]
        utils.record_list_to_csv(record_list=prior_list, output_filename='test_file')
        ### 
        added_list = [{'key1': 'value31', 'key2': 'value32', 'key3': 'value33', 'key4': 'value34'},
                           {'key1': 'value41', 'key2': 'value42', 'key3': 'value43'}] 
        utils.record_list_to_csv(record_list=added_list, output_filename='test_file', append=True)
        output_list = []
        with open('test_file.csv', 'r') as r:
            reader = csv.DictReader(r)
            for row in reader:
                output_list.append(row)
        expected_list = [{'key1': 'value11', 'key2': 'value12', 'key3': 'value13', 'key4': ''},
                         {'key1': 'value21', 'key2': 'value22', 'key3': 'value23', 'key4': ''},
                         {'key1': 'value31', 'key2': 'value32', 'key3': 'value33', 'key4': 'value34'},
                         {'key1': 'value41', 'key2': 'value42', 'key3': 'value43', 'key4': ''}]
        assert(output_list == expected_list)
        os.chdir('..')
        shutil.rmtree('testing_folder')

    
    ### --- record_list_dict_to_csv tests
    @patch.object(utils, 'record_list_to_csv')
    def test_record_list_dict_to_csv(self, mock_record_list_to_csv):
        """
        #### Function: 
            - utils.record_list_dict_to_csv 
        #### Inputs: 
            -@record_list_dict: dict of lists, of dicts 
            -@filename_prefix: 'test_file'
        #### Expected Behaviour: 
            - for each key in the list, pass the value through to record_list_to_csv
        #### Assertions: 
            - record_list_to_csv is called the right number of times with the right parameters 
        """
        input_list_dict = {'key1': [{'first': 'call'}],
                           'key2': [{'second': 'call'}],
                           'key3': [{'third': 'call'}]}
        utils.record_list_dict_to_csv(input_list_dict, 'test_file')
        expected_calls = [call(records_list=[{'first': 'call'}], output_filename='test_file_key1.csv', append=False),
                            call(records_list=[{'second': 'call'}], output_filename='test_file_key2.csv', append=False),
                            call(records_list=[{'third': 'call'}], output_filename='test_file_key3.csv', append=False)]
        mock_record_list_to_csv.assert_has_calls(expected_calls)


    ### --- combine_records tests --- 
    def test_combine_records(self):
        """
        #### Function: 
            - uils.combine_records
        #### Inputs: 
            -@record_one: a dict
            -@record_two, a dict with no overlapping keys from record_one
        #### Expected Behaviour: 
            - for each key in record_two, it's not found in the return_record, so 
                it gets added to it with the value, 
            - the return_record is returned
        #### Assertions: 
            - The expected dict is returned
        """
        record_one = {'key1' : 'value1', 'key2': 'value2', 'key3': 'value3'}
        record_two = {'key4': 'value4' , 'key5': 'value5', 'key6': 'value6'}
        resp = utils.combine_records(record_one, record_two)
        expected_result = {'key1': 'value1', 'key2' : 'value2', 'key3' : 'value3', 
                           'key4': 'value4', 'key5': 'value5', 'key6': 'value6'}
        assert resp == expected_result
    
    def test_combine_records_empty(self):
        """
        #### Function: 
            - utils.combine_records
        #### Inputs: 
            -@record_one: empty dict 
            -@record_two: empty dict 
        #### Expected Behaviour: 
            - there aren't any values to loop through, nothing is added to record_one, which is
                already empty, so it returns empty
        #### Assertions:
            - The returned dict is empty
        """
        assert utils.combine_records({}, {}) == {}

    def test_combine_records_single_collision_overwrite(self):
        """
        #### Function: 
            - utils.combine_records
        #### Inputs: 
            -@record_one: a dict
            -@record_two: a dict with a shared key with record_one
        #### Expected Behaviour: 
            - when the collision is found on key1, the value from record_two is discarded
        #### Assertions:
            - the returned dict is as expected, with no value4 in it 
        """
        record_one = {'key1' : 'value1', 'key2': 'value2', 'key3': 'value3'}
        record_two = {'key1': 'value4' , 'key5': 'value5', 'key6': 'value6'}
        resp = utils.combine_records(record_one, record_two)
        expected_result = {'key1': 'value1', 'key2' : 'value2', 'key3' : 'value3', 
                             'key5': 'value5', 'key6': 'value6'}
        assert resp == expected_result




        


import requests
import json
import urllib
from .utils import *

class Sftocsv:

    def __init__(self, base_url: str, api_version: float, access_token: str = '', tokenless: bool=False):
        """
        #### Inputs:
            -@base_url: Salesforce org url (i.e 'https://examplecompany.my.salesforce.com') 
            -@api_version: salesforce api version in float format (i.e 58.0)
            -@access_token: client credentials flow access token 
            -@tokenless: disables missing token exception. Useful if you want the joins and don't need to query
        """
        self.base_url = base_url
        self.api_version = f'v{str(api_version)}' ## 58.0
        if not access_token and not tokenless:
            raise Exception('Access Token missing. If you want to use non-query functions pass in tokenless = True')
        self.access_token = access_token 

  
    def query_records(self, querystring: str, nested: bool=False) -> list[dict] | dict[str, list[dict]]: 
        """
        #### Inputs:
            -@queryString: soql of form 'SELECT ... from ... where ...'
            -@nested: If you're using this function for a nested query, set to True 
        #### Expected Behaviour: 
            - the input string is url-parsed and the request is sent
            - the results are paginated through if required and records are read into a list of dicts 
            - if 'nested' is true, the 'attributes' section of each record is kept so that child records 
                can inheret a parent record id and save it in a field of the parent type. 
                If it's not nested, the attributes section is removed. 
        ##### Returns:
            - list[dict]: in the case of nested=False, returns a list of dicts, each dict being a record from the query 
            - dict[str, list[dict]]: in the case of nested=True, list of records are stored against a key of their type in the dict
        #### Side Effects:  
            - None 
        #### Exceptions: 
            - If status_code returned by query != 200, re-raises the error as an exception
        """
        querystring = urllib.parse.quote_plus(querystring)
        urlstring = f"{self.base_url}/services/data/{self.api_version}/query/?q={querystring}"
        header_dict = {"Authorization": f"Bearer {self.access_token}"}
        resp = requests.get(url=urlstring, headers=header_dict)
        if resp.status_code != 200:
            raise Exception(f'Query of -->{querystring}<-- raised error: \n {str(resp.content)}')
        resp_json = json.loads(resp.content)
        records = resp_json['records']
        next_url = resp_json.get('nextRecordsUrl', None)
        while next_url: 
            resp = requests.get(url=f"{self.base_url}{next_url}", headers=header_dict)
            if resp.status_code != 200:
                raise Exception(f'Query of -->{querystring}<-- on nextUrl -->{next_url}<-- raised error: \n {str(resp.content)}')
            resp_json = json.loads(resp.content)
            records += resp_json['records']
            next_url = resp_json.get('nextRecordsUrl', None)
        if(not nested): 
            for record in records:
                del(record['attributes'])
        else: 
            records = utils.split_nested_record_list(records)
        return records
            
            
    def large_in_query(self, querystring: str, in_list:list, nested: bool=False) -> list[dict] | dict[str, list[dict]]: 
        """
        #### Inputs:
            -@querystring: soql of form 'SELECT ... from ... where ... <in> ...'
            -@in_list: a list to substitute into the spot of <in> 
            -@nested: If you're using this function for a nested query, set to True
        #### Expected Behaviour: 
            - This is for use in avoiding hitting the 20,000 character limit on a query. You're likely to only hit this if you have a 
                very large list of values in a 'in' query. This function splits the 'in_list' into amounts that will fit into the 
                query limit, sends the requests and then combines the results and returns.
            - if the querystring doesn't contain a '<in>' substring an exception will be thrown
        #### Returns:   
            - list[dict]: Returned in the case of nested=True; a list of dicts, each dict being a record from the query 
            - dict[str, list[dict]]: Returned in the case of nested=False; list of records are stored against a key of their type in the dict
        #### Side Effects: 
            - None 
        #### Exceptions:
            - No '<in>' found...: Raised if the querystring has no <in> substring
            - in_list is empty: Raised if in_list is empty
        """
        if('<in>' not in querystring):
            raise Exception(f'No <in> found in query -->{querystring}<--')
        if(len(in_list) == 0):
            raise Exception('in_list is empty')
        
        built_querystring, remaining_list = utils.build_in_querystring(querystring, in_list)
        current_records = [] 
        if(nested):
            current_records = {}
        while len(remaining_list) > 0:
            resp = self.query_records(built_querystring)
            if(nested):
                current_records = utils.combine_nested_result_dicts(source_dict=resp, destination_dict=current_records)
            else:
                current_records += resp
            built_querystring, remaining_list = utils.build_in_querystring(querystring, remaining_list)
        resp = self.query_records(built_querystring)
        if(nested):
            current_records = utils.combine_nested_result_dicts(source_dict=resp, destination_dict=current_records)
        else:
            current_records += resp
        return(current_records)


    @staticmethod
    def records_to_csv(records: list[dict] | dict[str, list[dict]] , output_filename: str, append: bool=False): #tested #need to make this work for nested as well 
        """
        #### Inputs: 
            -@records: either a list of dicts (representing a non-nested query result), 
                        of a dict with lists of dicts as values, (representing a nested query result)
            -@output_filename: string to use as the filename. (.csv format is optional on the end) 
                In the case of a nested query result this string will become a prefix and the record type will be appended (i.e _Account.csv)
            -@append: if True, will attempt to append to the file, rather than write new ones
        #### Expected Behaviour: 
            - First checks that .csv wasn't passed in, trims it if it was, so that it can be used as a 
                prefix if needed in record_list_dict_to_csv
            - then depending on the type of records parameter passed in, either calls 
                utils.record_list_to_csv or record_list_dict_to_csv 
            
        Takes in a list of records (dicts) and a filename ending with csv,
        Each record will be saved as a row in the csv output
        Each unique key will become a header in the csv
        """
        print(type(records))
        output_filename = output_filename.rpartition('.csv')[0]
        if(type(records) == list):
            utils.record_list_to_csv(record_list=records, output_filename=output_filename, append=append)
        elif(type(records) == dict):
            utils.record_list_dict_to_csv(record_list_dict=records, filename_prefix=output_filename, append=append)

    
    @staticmethod
    def inner_join(left_list: list[dict], right_list: list[dict], left_key: str, right_key:str, preserve_right_key:bool=False) -> list[dict]:
        """
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict]
            -@left_key: the key you want to match with from the left_list
            -@right_key: the key you want to match with from the right_lis
            -@preserve_right_key: If true, will keep the right key in the resulting dicts, otherwise 
                just preserves the left key 
        #### Expected Behaviour: 
            -will produce a list of records equivalent to an INNER JOIN 
                (exclusively rows that have a key found in both lists are combined and output)
            - i.e, for each value in left_list, if left_key matches any record in right_list on the right_key, 
                those records are combined into the output. 
        #### Returns: 
            - The resulting dict
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None 
        """
        resulting_list = []
        for left_record in left_list:
            if left_key not in left_record.keys():
                continue
            for right_record in right_list:
                if right_key not in right_record.keys():
                    continue
                if right_record[right_key] == left_record[left_key]:
                    combined_record = utils.combine_records(left_record, right_record)
                    if not preserve_right_key:
                        del combined_record[right_key]
                    resulting_list.append(combined_record)
        return(resulting_list)
    

    @staticmethod
    def natural_join(left_list: list[dict], right_list: list[dict], exclusive:bool=False):     
        """
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict]
            -@exclusive: If False -> any single matching column name + value will create a join
                         If True -> all shared column names must match on value for a row to join 
        #### Expected Behaviour: 
            - for each record in both: 
                - if it's an inclusive method (exclusive = False), it looks for any key that is shared and 
                    if two records share a value they are combined and returned. Any matching keys that don't 
                    share a value wil result in having the right_list value  
                - if it's an exclusive method (exclusive = True), it looks for records in which all shared columns 
                    share values 
            - Return combined list result
        #### Returns: 
            - List of dicts resulting from join
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - None 
        """
        resulting_list = []
        for left_record in left_list:
            for right_record in right_list:
                match = False
                retrieved = [left_record.get(key) == right_record.get(key) for key in left_record.keys() if key in right_record.keys()]
                #inclusive method, if any of these match it's all good 
                if not exclusive and True in retrieved:  
                    match = True                        
                #exclusive method, all must match on shared keys 
                elif exclusive and False not in retrieved:
                    match = True 
                if(match == True):                    
                    combined_record = utils.combine_records(left_record, right_record)
                    resulting_list.append(combined_record)
                    break
        return(resulting_list)
    

    @staticmethod
    def outer_join(left_list: list[dict], right_list: list[dict], left_key: str, right_key: str,
                    side: str, preserve_innner_key:bool=False):
        """
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict]
            -@left_key: key to match upon from the left_list 
            -@right_key: key to match upon from the right_list
            -@side: on of ['left', 'right', 'full'], to designate the type of outer join
        #### Expected Behaviour: 
            - if the 'side' is entered, it fills the outer, inner, outer_key, inner_key accordingly 
                for each record in the outer list, it is checked against all elements of the inner list,
                anny inner record that matches results in a new combined record being created and added to the return list
            - if an outer record is not matched on any inner record, it is still appended to the return list 
            - for the inner list, a set of indexes of records that are unmatched is kept. In the case of a full outer join,
                even if the inner list record is never matched against an outer record, it should still be added to the list. 
                if it is matched, it doesn't need adding to the list again, hence the unmatched_inner storage. 
        #### Returns: 
            - List of records resulting from the join 
        #### Side Effects: 
            - None 
        #### Exceptions: 
            - 'outer_join requires one of..' If the 'side' doesn't match one of the valid values 
        """
        side_map = {'left': {'outer' : left_list, 'outer_key': left_key, 'inner': right_list, 'inner_key': right_key},
                    'right': {'outer':right_list, 'outer_key':  right_key, 'inner': left_list, 'inner_key': left_key},
                    'full': {'outer' : left_list, 'outer_key': left_key, 'inner': right_list, 'inner_key': right_key}}
        if side not in side_map.keys():
            raise Exception('outer_join requires one of ("left", "right", "full") in "side" argument')
       
        #outer and inner 
        outer_key = side_map[side]['outer_key']
        outer = side_map[side]['outer']
        inner_key = side_map[side]['inner_key']
        inner = side_map[side]['inner']
        return_list = []
        unmatched_inner = set(list(range(len(inner))))
        for outer_record in outer: 
            matched = False
            for inner_i, inner_record in enumerate(inner):
                if outer_record.get(outer_key) == inner_record.get(inner_key):
                    matched = True
                    combined_record = utils.combine_records(outer_record, inner_record)
                    if(not preserve_innner_key):
                        del(combined_record[inner_key])
                    return_list.append(combined_record)
                    if(inner_i in unmatched_inner):
                        unmatched_inner.remove(inner_i)
            if not matched:
                return_list.append(outer_record)
        if side == 'full':
            for x in unmatched_inner:
                return_list.append(inner[x])
        return(return_list)



from sftocsv import Sftocsv
from sftocsv import utils
from unittest.mock import Mock, patch, call 
import requests
import json
from unittest import TestCase
import os
import csv
class test_sftocsv(TestCase):
    if os.path.isfile("/tmp/sf_token_store.json"):
        os.remove('/tmp/sf_token_store.json') #this can't be kept forever. Tests should never affect production

    ### --- __init__ tests --- 
    def test_init_no_access_token(self):
        """
        #### Function: 
            - Sftocsv.__init__
        #### Inputs: 
            -@base_url: 'https://examplecompany.my.salesforce.com'
            -@api_version: 58.0
            -@access_token: Not passed in 
            -@tokenless: not passed in
        #### Expected Behaviour:    
            - Because no access_token is passed in and tokenless isn't True, an exception is raised
        #### Assertions: 
            - the expected exception is thrown
        """
        with self.assertRaises(Exception) as context: 
            Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0)
        assert(str(context.exception) == 'Access Token missing. If you want to use non-query functions pass in tokenless = True')

    def test_init(self):
        """
        #### Function:
            - Sftocsv.__init__
        #### Inputs: 
            -@base_url: 'https://examplecompany.my.salesforce.com'
            -@api_version: 58.0
            -@access_token: 'test_token'
            -@tokenless: not passed in 
        #### Expected Behaviour: 
            - Because access_token is passed in, all attributes are set and no exceptions are thrown
        #### 
            - the expected attributes are set on the instance of Sftocsv
        """
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        assert(resource.base_url == 'https://examplecompany.my.salesforce.com')
        assert(resource.api_version == 'v58.0')
        assert(resource.access_token == 'test_token')


    ### --- query_records tests ---
    @patch("requests.get")
    def test_query_records_single_200(self, mock_get):
        """
        #### Function: 
            - Sftocsv.query_records
        #### Inputs:
            -@querystring: 'select id, field from opportunity' (request is patched so string not important)
            -@nested: False
        #### Expected Behaviour: 
            - The resp.status_code is mocked to 200 so the json is loaded in,
                the next_url is not found, and nested is false, so attributes are removed
                from each record
            - the records are returned
        #### Assertions: 
            - the returned records are the same as the mocked data, minus the 'attributes' sections in each
        """
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        mock_response = requests.Response()
        mock_response._content = b"""{"totalSize":1,"done":true,"records":[{"attributes":{"type":"Opportunity","url":"/services/data/v58.0/sobjects/Opportunity/fake_id"},"Id":"Id1", "field": "value1"},
                                                                         {"attributes":{"type":"Opportunity","url":"/services/data/v58.0/sobjects/Opportunity/fake_id"},"Id":"Id2", "field": "value2"}]}"""
        mock_response.status_code = 200
        mock_get.return_value = mock_response
        resp = resource.query_records("select id, field from opportunity", nested=False)
        expected_value = [{"Id" : "Id1", "field": "value1"}, {"Id": "Id2", "field": "value2"}]
        assert resp == expected_value

    @patch("requests.get")
    def test_query_records_error(self, mock_get):
        """
        #### Function: 
            - Sftocsv.query_records
        #### Inputs:
            -@querystring: 'select id, field from opportunity'
            -@nested: False
        #### Expected Behaviour: 
            - The mocked response is non-200 so an exception is raised and the querystring is
                present in the error message 
        #### Assertions: 
            - the expected exception is raised
        """
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        mock_get_response = requests.Response()
        mock_get_response.status_code = 400
        mock_get_response._content = "error"
        mock_get.return_value = mock_get_response
        with self.assertRaises(Exception) as context:
            resource.query_records("select id, field from opportunity")
        assert(f'Query of -->select+id%2C+field+from+opportunity<-- raised error: \n error' == str(context.exception)) 

    @patch("requests.get")
    def test_query_records_empty(self, mock_get):
        """
        #### Function: 
            - Sftocsv.query_records
        #### Inputs: 
            -@querystring: 'select id, field from opportunity
            -@nested: False
        #### Expected Behaviour:   
            - The mocked response is an empty list of records, so an empty list of records is returned
        #### Assertions: 
            - The returned list is empty
        """
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        mock_get_response = requests.Response()
        mock_get_response.status_code = 200
        mock_get_response._content = b'{"totalSize":1,"done":true,"records":[]}'
        mock_get.return_value = mock_get_response
        resp = resource.query_records('select id, field from opportunity')
        assert resp == []

    @patch("requests.get")
    def test_query_records_long(self, mock_get):
        """
        #### Function: 
            - Sftocsv.query_records
        #### Inputs: 
            -@querystring: 'select id from opportunity' 
            -@nested: False
        #### Expected Behaviour: 
            - The first mocked response has a nextRecordsUrl so another request is sent, 
                same with the second mocked response 
            - The third mocked response has no nextRecordsUrl so the requests finish, 
            - The retrieved records from each call are collated and returned
        Assert that when a nextRecordsUrl is returned it will loop through to collect them all 
        """
        response_1 = requests.Response()
        response_1.status_code = 200
        response_1._content =  b'{"totalSize":1,"done": false, "nextRecordsUrl": "fake_url", "records":[{"attributes":{"type":"Opportunity","url":"/services/data/v58.0/sobjects/Opportunity/fake_id"},"Id":"Id1"}]}'
        response_2 = requests.Response()
        response_2.status_code = 200
        response_2._content = b'{"totalSize":1,"done": false, "nextRecordsUrl": "fake_url", "records":[{"attributes":{"type":"Opportunity","url":"/services/data/v58.0/sobjects/Opportunity/fake_id"},"Id":"Id2"}]}'
        response_3 = requests.Response()
        response_3.status_code = 200
        response_3._content = b'{"totalSize":1,"done": true, "records":[{"attributes":{"type":"Opportunity","url":"/services/data/v58.0/sobjects/Opportunity/fake_id"},"Id":"Id3"}]}'
        mock_get.side_effect = [response_1, response_2,response_3]
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        resp = resource.query_records('select id from opportunity')
        expected_result = [{'Id': 'Id1'}, {'Id' : 'Id2'}, {'Id': 'Id3'}]
        assert resp == expected_result

    @patch("requests.get")
    def test_query_records_long_error(self, mock_get):
        """
        #### Function: 
            - Sftocsv.query_records
        #### Inputs: 
            -@querystring: 'select id from opportunity' 
            -@nested: False
        #### Expected Behaviour: 
            - The first mocked response has a nextRecordsUrl so another request is sent, 
                same with the second mocked response
            - The third mocked response has a 400 status code, so an exception is thrown with the content of the message
        #### Assertions: 
            - The expected Exception is thrown
        """
        response_1 = requests.Response()
        response_1.status_code = 200
        response_1._content =  b'{"totalSize":1,"done": false, "nextRecordsUrl": "fake_url", "records":[{"attributes":{"type":"Opportunity","url":"/services/data/v58.0/sobjects/Opportunity/fake_id"},"Id":"Id1"}]}'
        response_2 = requests.Response()
        response_2.status_code = 200
        response_2._content = b'{"totalSize":1,"done": false, "nextRecordsUrl": "fake_url", "records":[{"attributes":{"type":"Opportunity","url":"/services/data/v58.0/sobjects/Opportunity/fake_id"},"Id":"Id2"}]}'
        response_3 = requests.Response()
        response_3.status_code = 400
        response_3._content = 'error'
        mock_get.side_effect = [response_1, response_2,response_3]
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        with self.assertRaises(Exception) as context:
            resp = resource.query_records('select id from opportunity')
        assert(f'Query of -->select+id+from+opportunity<-- on nextUrl -->fake_url<-- raised error: \n error' == str(context.exception)) 

    @patch.object(utils, 'split_nested_record_list')
    @patch("requests.get")
    def test_query_records_nested(self, mock_get, mock_split_nested_record_list):
        """
        #### Function: 
            - Sftocsv.query_records
        #### Inputs: 
            -@querystring: 'select name, (select lastname from contacts) from account'
            -@nested: True
        #### Expected Behaviour: 
            - The resp.status_code is mocked to 200 so the json is loaded in, the next_url is not found,
            - nested is true, so attributes are not removed from each record
            - split_nested_record is called on the response records
            - the dict returned by split_nested_recored is returned
        #### Assertions: 
            - split_nested_record_list is called once with the response from the query 
        """
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        mock_response = requests.Response()
        mock_response._content = b'{"totalSize": 1, "done": true, "records": [{"attributes":{"type":"Account","url":"contact_url"},"Name":"Test Name","Id":"test_id","Contacts":{"totalSize":1,"done":true,"records":[{"attributes":{"type":"Contact","url":"contact_url"},"AccountId":"test_a_id","Id":"t_id","LastName":"Test"}]}}]}'
        mock_response.status_code = 200
        mock_get.return_value = mock_response
        resp = resource.query_records("select name, (select lastname from contacts) from account", nested=True)
        expected_value = [{"attributes":{"type":"Account","url":"contact_url"},"Name":"Test Name","Id":"test_id","Contacts":{"totalSize":1,"done":True,"records":[{"attributes":{"type":"Contact","url":"contact_url"},"AccountId":"test_a_id","Id":"t_id","LastName":"Test"}]}}]
        mock_split_nested_record_list.assert_called_once_with(expected_value)


    ### --- large_in_query tests --- 
    def test_large_in_query_errors(self):
        """
        #### Function: 
            - Sftocsv.large_in_query
        #### Inputs: 
            2 calls
            1. -@querystring: 'select id from opportunity' (has no <in> in it )
                -@in_list: list of size one 
            2. -@querystring: 'select name from opportunity where id in <in> 
                -@in_list: empty list
        #### Expected Behaviour: 
            - One the first call the condition regarding missing <in> substring is true, so an exception is thrown
            - on the second call the condition regarding the empty in_list list is true, so an exception is thrown 
        #### Assertions: 
            - The expected exception is raised in each case 
        """
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        with self.assertRaises(Exception) as context:
            resp = resource.large_in_query(querystring='select id from opportunity', in_list=['testing'])
        assert(str(context.exception) == "No <in> found in query -->select id from opportunity<--")
        with self.assertRaises(Exception) as context:
            resp = resource.large_in_query(querystring='select name from oppportunity where id in <in>', in_list=[])
        assert(str(context.exception) == 'in_list is empty')

    @patch('sftocsv.Sftocsv.query_records')
    @patch.object(utils, 'build_in_querystring')
    def test_large_in_query_not_nested(self, mock_build_in_querystring, mock_query_records):
        """
        #### Function: 
            - Sftocsv.large_in_query
        #### Inputs: 
            -@querystring 'select id from opportunity where id in <in>
            -@in_list: non empty list (both not important as it's mocked)
        #### Expected Behaviour:
            - The first call to build_in_querystring returns a remaining_list that is non empty,
                so the while loop is entered. query_records returns a mocked response which is appended to
                the current_records list 
            - the second call to build_in_querystring returns an empty remaining_list,
            - the second call to query_records returns a different set of mocked records, these are appended to the current_records
            - current_records is returned
        #### Assertions: 
            - Build_in_querystring is called twice, with the expected reduction in the remaining_list between them 
            - query_records is called with the expected querystring returned by build_in_querystring
        """
        first_build = ("select id from opportunity where id in ('Id1', 'Id2', 'Id3')", ['Id4', 'Id5', 'Id6'])
        second_build = ("select id from opportunity where id in ('Id4', 'Id5', 'Id6')", [])
        mock_build_in_querystring.side_effect = [first_build, second_build]

        first_result = [{'Id': 'Id1'}, {'Id' : 'Id2'}, {'Id': 'Id3'}]
        second_result = [{'Id': 'Id4'}, {'Id' : 'Id5'}, {'Id': 'Id6'}]
        mock_query_records.side_effect = [first_result, second_result]
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        resp = resource.large_in_query('select id from opportunity where id in <in>', ['Id1', 'Id2', 'Id3', 'Id4', 'Id5', 'Id6'])
        assert resp == [{'Id': 'Id1'}, {'Id' : 'Id2'}, {'Id': 'Id3'},{'Id': 'Id4'}, {'Id' : 'Id5'}, {'Id': 'Id6'}]
        expected_build_calls = [call('select id from opportunity where id in <in>', ['Id1', 'Id2', 'Id3', 'Id4', 'Id5', 'Id6']),
                                call('select id from opportunity where id in <in>', ['Id4', 'Id5', 'Id6'])]
        mock_build_in_querystring.assert_has_calls(expected_build_calls)
        mock_query_records.assert_has_calls([call("select id from opportunity where id in ('Id1', 'Id2', 'Id3')"), 
                                             call("select id from opportunity where id in ('Id4', 'Id5', 'Id6')")])
        
    @patch('sftocsv.Sftocsv.query_records')
    @patch.object(utils, 'build_in_querystring')
    def test_large_in_query_nested(self, mock_build_in_querystring, mock_query_records):
        """
        #### Function: 
            - Sftocsv.large_in_query
        #### Inputs:    
            -@querystring: 'select id from opportunity where id in <in>
            -@in_list: non empty list (both not important as it's mocked)
            -@nested: True
        #### Expected Behaviour: 
            - The first call to build_in_querystring returns a remaining_list that is not empty, 
                so the while loop is entered. query_records returns a mocked response that is combined into current_records
            - the second call to the build_in_querystring returns an empty remaining_list 
            - the second call to query_records returns a different set of mocked records, these are also combined into the current_records
            - current_records is returned as expected
        #### Assertions: 
            - build_in_querystring is called twice, with the expected retuduction in the remaining_list between them 
            - query_records is called with the expected querystring returned by build_in_querystring 
            - the returned records are as expected
        """
        first_build = ("select id from opportunity where id in ('Id1', 'Id2', 'Id3')", ['Id4', 'Id5', 'Id6'])
        second_build = ("select id from opportunity where id in ('Id4', 'Id5', 'Id6')", [])
        mock_build_in_querystring.side_effect = [first_build, second_build]

        first_response = {'Account': [{'Name': 'Test Name', 'Id': 'test_id'}],
                          'Contact': [{'AccountId': 'test_a_id', 'Id': 't_id', 'LastName': 'Test'}]}
        second_response = {'Account': [{'Name': 'Test Name2', 'Id': 'test_id2'}],
                           'Contact': [{'AccountId': 'test_a_id2', 'Id': 't_id2', 'LastName': 'Test2'}]}
        mock_query_records.side_effect = [first_response, second_response]
        resource = Sftocsv(base_url='https://examplecompany.my.salesforce.com', api_version=58.0, access_token='test_token')
        resp = resource.large_in_query('select id from opportunity where id in <in>', ['Id1', 'Id2', 'Id3', 'Id4', 'Id5', 'Id6'], nested=True)
        expected_resp = {'Account': [{'Name': 'Test Name', 'Id': 'test_id'}, {'Name': 'Test Name2', 'Id': 'test_id2'}], 
                         'Contact': [{'AccountId': 'test_a_id', 'Id': 't_id', 'LastName': 'Test'}, {'AccountId': 'test_a_id2','Id': 't_id2', 'LastName': 'Test2'}]}
        assert(resp == expected_resp)
        mock_build_in_querystring.assert_has_calls([call('select id from opportunity where id in <in>', ['Id1', 'Id2', 'Id3', 'Id4', 'Id5', 'Id6']),
                                                    call('select id from opportunity where id in <in>', ['Id4', 'Id5', 'Id6'])])
        mock_query_records.assert_has_calls([call("select id from opportunity where id in ('Id1', 'Id2', 'Id3')"), 
                                             call("select id from opportunity where id in ('Id4', 'Id5', 'Id6')")])

    ### --- records_to_csv tests ---
    @patch.object(utils, 'record_list_dict_to_csv')
    @patch.object(utils, 'record_list_to_csv')
    def test_records_to_csv(self, mock_record_list_to_csv, mock_record_list_dict_to_csv):
        """
        #### Function: 
            - Sftocsv.records_to_csv
        #### Inputs: 
            2 calls
            1. - a list of dicts (representing a non-nested result)
            2. - a dict of list (representing a nested result)
        #### Expected Behaviour: 
            - 1st call, record_list_to_csv is called 
            - 2nd call, record_list_dict_to_csv is called 
        #### Assertions: 
            -1st call record_list_to_csv is called 
            -2nd call record_list_dict_to_csv is called
        """
        Sftocsv.records_to_csv([{}], output_filename='test')
        mock_record_list_to_csv.assert_called_once()
        Sftocsv.records_to_csv({'try': [{}]}, output_filename='test')
        mock_record_list_dict_to_csv.assert_called_once()
    

    ### --- inner_join tests ---
    def test_inner_join_shared_key_single_no_preserve(self):
        """
        #### Function:
            - Sftocsv.inner_join
        #### Inputs:
            -@left_list: a list with dict inside 
            -@right_list: a list with dict inside, it shares a value in a column with left_list
            -@left_key: 'customer_id'
            -@right_key: 'cust_id'
            -@preserve_right_key: False
        #### Expected Behaviour: 
            - the dicts share a value on their respective keys, so they are combined,
            - then because the preserve_right_key is False, the right_key is deleted from 
                the resulting dict 
        #### Assertions:
            - The returned list[dict] is a combination of input_1's dict and input_2's dict, 
                and does not contain the right key
        """
        input_1 = [{"customer_id" : "value1", "key2": "value2", "key3": "value3"}]
        input_2 = [{"cust_id": "value1", "key4": "value4", "key5": "value5"}]
        resp = Sftocsv.inner_join(input_1, input_2, 'customer_id', 'cust_id')
        assert resp == [{"customer_id" : "value1", "key2": "value2", "key3": "value3", "key4" : "value4", "key5": "value5"}]
    
    def test_inner_join_shared_key_single_preserve(self):
        """
        ### Function:
            - Sftocsv.inner_join
        #### Inputs:
            -@left_list: a list with dict inside 
            -@right_list: a list with dict inside, it shares a value in a column with left_list
            -@left_key: 'customer_id'
            -@right_key: 'cust_id'
            -@preserve_right_key: True
        #### Expected Behaviour: 
            - the dicts share a value on their respective keys, so they are combined,
            - then because the preserve_right_key is True, the right_key is maintained in the result
        #### Assertions: 
            - The returned list[dict] is a combination of input_1's dict and input_2's dict, 
                and does contain the right key
        """
        input_1 = [{"customer_id" : "value1", "key2": "value2", "key3": "value3"}]
        input_2 = [{"cust_id": "value1", "key4": "value4", "key5": "value5"}]
        resp = Sftocsv.inner_join(input_1, input_2, 'customer_id', 'cust_id', preserve_right_key=True)
        assert resp == [{"customer_id" : "value1", "key2": "value2", "key3": "value3",
                          "cust_id": "value1", "key4" : "value4", "key5": "value5"}]

    def test_inner_join_empties(self):
        """
         ### Function:
            - Sftocsv.inner_join
        #### Inputs:
            2 calls
            -@left_list: 1st call, a list with a dict inside, second call empty
            -@right_list: 1st call, empty, 2nd call, a list with a dict inside
            -@left_key: 1st call 'customer_id' (found in left_list), 2nd call, 'cust_id' 
            -@right_key: 1st call 'cust_id', 2nd call 'customer_id' (found in right list)
            -@preserve_right_key: True
        #### Expected Behaviour: 
            - Because one dict is present in each case, there is no match on any key, so the returned dict is empty 
        #### Assertions: 
            - The returned list is empty both calls 
        Assert that when inner_join is passed a non-empty left list and an empty right list,
        it will return an empty list
        """
        example_dict = {"customer_id": "value1", "key4": "value4", "key5": "value5"}
        resp = Sftocsv.inner_join([example_dict], [], "customer_id", "cust_id")
        assert resp == []
        resp = Sftocsv.inner_join([], [example_dict], 'cust_id', 'customer_id')
        assert(resp == [])
    
    def test_inner_join_incorrect_keys(self):
        """
        #### Function: 
            - Sftocsv.inner_join
        #### Inputs: 
            2 calls 
            -@left_list: list with 1 dict element
            -@right_list: list with 1 dict element 
            -@left_key: 1st call, a key not found in the left list, 2nd call, a key found in the left list 
            -@right_key: 1st call, a key found in the right list, 2nd call, a key not found in the right list
        #### Expected Behaviour: 
            - On each call the inability to find the key means there is no way the associated values can match,
                so there is no join and the returned list is empty 
        #### Assertions:
            - The returned dict in each case is empty 
        """
        input_1 = [{"customer_id" : "value1", "key2": "value2", "key3": "value3"}]
        input_2 = [{"cust_id": "value1", "key4": "value4", "key5": "value5"}]
        resp = Sftocsv.inner_join(input_1, input_2, "other_id", "never_checked")
        assert resp == []
        resp = Sftocsv.inner_join(input_1, input_2, 'customer_id', 'not_id')
        assert resp == []
        
    def test_inner_join_duplicates(self):
        """
        #### Function: 
            - Sftocsv.inner_join
        #### Inputs: 
            -@left_list: a list with duplicate values in the left_key
            -@right_list: a list with duplicate values in the right_key
            -@left_key: value found in left_list
            -@right_key: value found in right_list
        #### Expected Behaviour: 
            - Because there are 2 rows matching 2 rows on the left/right key, they combine to produce 4 rows
        #### Assertions: 
            - The returned list has 4 rows 
        """
        input_1 = [
            {"customer_id": "value1", "key2": "value2", "key3": "value3"},
            {"customer_id": "value1" , "key4" : "value4", "key5": "value5"}
        ]
        input_2 = [
            {"cust_id": "value1", "key6": "value6", "key7": "value7"},
            {"cust_id": "value1", "key8": "value8", "key9": "value9"}
        ]
        resp = Sftocsv.inner_join(input_1, input_2, "customer_id", "cust_id")
        expected_response = [
            {'customer_id': 'value1', 'key2': 'value2', 'key3': 'value3', "key6": "value6", "key7": "value7"},
            {'customer_id': 'value1', "key2": "value2", "key3": "value3", "key8": "value8", "key9": "value9"},
            {'customer_id': 'value1', "key4": "value4", "key5": "value5", "key6": 'value6', 'key7': 'value7'},
            {'customer_id': 'value1', 'key4': 'value4', 'key5': 'value5', 'key8': 'value8', 'key9' : 'value9'}
        ]
        assert resp == expected_response


    ### --- natural_join tests ---
    def test_natural_join_empties(self):
        """
        #### Function: 
            - Sftocsv.natural_join
        #### Inputs: 
            3 calls
            -@left_list: 1st call, empty, 2nd call, non-empty, 3rd call, empty
            -@right_list: 1st call, non-empty, 2nd all, empty, 3rd call, empty
        #### Expected Behaviour: 
            - In each call, because an empty list cannot have any fields to join on the resulting list should be empty
        #### Assertions: 
            - Each response is an empty list
        """
        resp = Sftocsv.natural_join([], [{'customer_id': 'id1', 'key1': 'value1'}])
        assert resp == []
        resp = Sftocsv.natural_join([{'customer_id': 'id1', 'key1': 'value1'}], [])
        assert resp == []
        resp = Sftocsv.natural_join([],[])
        assert resp == []

    #### --- Inclusive section 
    def test_natural_join_single_match_inclusive(self):
        """
        #### Function:
            - Sftocsv.natural_join
        #### Inputs: 
            -@left_list: list[dict] with one column name shared with matching value with right_list 
            -@right_list: list[dict] with one column name shared with matching value with left list
            -@exclusive: False
        #### Expected Behaviour: 
            - These records have one shared column and it has a matching value, so they are combined and added to the returned list 
        #### Assertions: 
            - The returned list has the one combined result in it 
        """
        input_1 = [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "matched_value"}]
        input_2 = [{'cust_id': 'id1', "key3" : "value3", "key4" : "value4", 'matched_key': "matched_value"}]
        resp = Sftocsv.natural_join(input_1, input_2)
        assert resp == [{'customer_id': 'id1', 'key1': "value1", 'key2' : 'value2', 'matched_key': 'matched_value',
                         'cust_id': 'id1', 'key3': 'value3', 'key4': 'value4'}]
    
    def test_natural_join_single_miss_inclusive(self):
        """
        #### Function: 
            - Sftocsv.natural_join
        #### Inputs: 
            -@left_list: list[dict] with no shared columns with right_list
            -@right_list: list[dict] with no shared columns with left_list 
        #### Expected Behaviour: 
            - These records share no column names, no attempt to combine is made, returned list is never 
                added to, it's empty when returned 
        #### Assertions: 
            - The returned list is empty 
        """
        input_1 =  [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "unmatched_value"}]
        input_2 = [{'cust_id': 'id1', "key3" : "value3", "key4" : "value4", 'matched_key': "other_value"}]
        resp = Sftocsv.natural_join(input_1, input_2)
        assert resp == []
         
    def test_natural_join_multi_match_inclusive(self):
        """
        #### Function: 
            - Sftocsv.natural_join
        #### Inputs:
            -@left_list: list[dict], 1 item with 2 column names shared with matching values with right list
            -@right_list: list[dict], 1 item with 2 column names shared with matching values with left list
            -@exclusive: False
        #### Expected Behaviour: 
            - These records have 2 shared columns and the values in these columns match, because it's inclusive only a single value 
                needs to match, so as soon as the matching value is found they're combined and added to the returned list 
        #### Assertions: 
            - The returned list has one combined result in it 
        """
        input_1 = [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "matched_value", 'matched_key_2': 'matched_value2'}]
        input_2 = [{'cust_id': 'id1', "key3" : "value3", "key4" : "value4", 'matched_key': "matched_value", 'matched_key_2': 'matched_value2'}]
        resp = Sftocsv.natural_join(input_1, input_2)
        assert resp == [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "matched_value", 'matched_key_2': 'matched_value2',
                         'cust_id': 'id1', "key3" : "value3", "key4" : "value4"}]


    #### --- Exclusive section 
    def test_natural_join_single_match_exclusive(self):
        """
        #### Function: 
            - Sftocsv.natural_join
        #### Inputs: 
            -@left_list: list[dict], 1 item with 1 column name shared with right list, the value matches in this column
            -@right_list: list[dict] 1 item with 1 column name shared with left list, the value matches in this column 
            -@exclusive: True
        #### Expected Behaviour: 
            - Because it's running in exclusive mode, all matching columnn names are found and it's checked that the values 
                match across these columns. They do, so it combines the records and adds the resulting record to the returned list 
        #### Assertions:
            - the returned list has one combined result in it
        """
        input_1 = [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "matched_value"}]
        input_2 = [{'cust_id': 'id1', "key3" : "value3", "key4" : "value4", 'matched_key': "matched_value"}]
        resp = Sftocsv.natural_join(input_1, input_2, exclusive=True)
        assert resp == [{'customer_id': 'id1', 'cust_id': 'id1', 'key1': "value1", 'key2' : 'value2', 'matched_key': 'matched_value',
                         'cust_id': 'id1', 'key3': 'value3', 'key4': 'value4'}]
    
    def test_natural_join_single_miss_exclusive(self):
        """
        #### Function: 
            - Sftocsv.natural_join
        #### Inputs: 
            -@left_list: list[dict] with no shared columns with right_list
            -@right_list: list[dict] with no shared columns with left_list 
            -@exclusive: True
        #### Expected Behaviour: 
            - These records share no column names, no attempt to combine is made, returned list is never 
                added to, it's empty when returned. It being exclusive has no effect on this 
        #### Assertions: 
            - The returned list is empty 
        """
        input_1 =  [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "unmatched_value"}]
        input_2 = [{'cust_id': 'id1', "key3" : "value3", "key4" : "value4", 'matched_key': "other_value"}]
        resp = Sftocsv.natural_join(input_1, input_2, exclusive=True)
        assert resp == []
        
    def test_natural_join_multi_match_exclusive(self):
        """
        #### Function: 
            - Sftocsv.natural_join
        #### Inputs:
            -@left_list: list[dict], 1 item with 2 column names shared with matching values with right list
            -@right_list: list[dict], 1 item with 2 column names shared with matching values with left list
            -@exclusive: True
        #### Expected Behaviour: 
            - These records have 2 shared columns and the values in these columns match, because it's Exclusive it's checked that 
                all values in these shared columns match, they do, so the records are combined and the combined record is   
                added to the returned list 
        #### Assertions: 
            - The returned list has one combined result in it 
        """
        input_1 = [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "matched_value", 'matched_key_2': 'matched_value2'}]
        input_2 = [{'cust_id': 'id1', "key3" : "value3", "key4" : "value4", 'matched_key': "matched_value", 'matched_key_2': 'matched_value2'}]
        resp = Sftocsv.natural_join(input_1, input_2, exclusive=True)
        assert resp == [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "matched_value", 'matched_key_2': 'matched_value2',
                         'cust_id': 'id1', "key3" : "value3", "key4" : "value4"}]
    
    def test_natural_join_multi_semimatch_exclusive(self):
        """
        #### Function:
            - Sftocsv.natural_join
        #### Inputs: 
            -@left_list: list[dict], 1 item with 2 column names shared with right_list, only 1 value shared
            -@right_list: list[dict], 1 item with 2 column names shared with left_list, only 1 value shared
            -@exclusive: True
        #### Expected Behaviour: 
            - These records have 2 shared column names, not all of these values match in the shared columns, 
                so the records are not combined, the returned list is empty
        #### Assertions: 
            - the returned list is empty  
        """
        input_1 = [{'customer_id': 'id1', "key1": "value1", "key2": "value2", 'matched_key': "matched_value", 'matched_key_2': 'no_match'}]
        input_2 = [{'cust_id': 'id1', "key3" : "value3", "key4" : "value4", 'matched_key': "matched_value", 'matched_key_2': 'not_here'}]
        resp = Sftocsv.natural_join(input_1, input_2, exclusive=True)
        assert resp == []   


    ### --- outer join tests --- 
    def test_outer_join_exception(self):
        """
        #### Function: 
            - Sftocsv.outer_join
        #### Inputs: 
            -@left_list: empty (not necessary)
            -@right_list: empty,
            -@left_key: 'test key'
            -@right_key: 'test key'
            -@side: 'other'
        #### Expected Behaviour: 
            - Because the entered 'side' value is not recognised an exception is thrown 
        #### Assertions: 
            - The expected exception is thrown 
        """
        with self.assertRaises(Exception) as context:
            resp = Sftocsv.outer_join([], [], 'test key', 'test key', 'other')
        assert(str(context.exception) == 'outer_join requires one of ("left", "right", "full") in "side" argument')
    
    def test_outer_join_left(self):
        """
        #### Function: 
            - Sftocsv.outer_join
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict]
            -@left_key: 'lkey' 
            -@right_key: 'rkey' 
            -@side: 'left'
            -@preserve_inner_key: False
        #### Expected Behaviour: 
            - The left_list is selected as the outer list, the right_list as the inner. 
            - the lists are looped through, the first record in the outer list matches the second in the inner, 
                these are combined and added the returned list. 
            - the second record in the outer list doesn't match anything, so it's just added the resulting list 
            - the third record in the outer matches 2 records in the inner, this results in 2 records added to the resulting list 
            - because preserve_inner_key is False, the inner matching key is not kept 
        #### Assertions: 
            - The returned list is as expected
        """
        input_left = [{'key1': 'val1a', 'key2': 'val2a', 'lkey': 'a'},
                      {'key1': 'val1b', 'key2': 'val2b', 'lkey': 'b'},
                      {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c'}]
        input_right = [{'key3': 'val1d', 'key4': 'val2d', 'rkey': 'd' },
                       {'key3': 'val1e', 'key4': 'val2e', 'rkey': 'a'},
                       {'key3': 'val1f', 'key4': 'val2f', 'rkey': 'c'},
                       {'key3': 'val1g', 'key4': 'val2g', 'rkey': 'c'}] 
        resp = Sftocsv.outer_join(left_list=input_left, right_list=input_right, 
                                  left_key='lkey', right_key='rkey', side='left')
        expected_result = [{'key1': 'val1a', 'key2': 'val2a', 'lkey': 'a', 'key3': 'val1e', 'key4': 'val2e'},
                           {'key1': 'val1b', 'key2': 'val2b', 'lkey': 'b'},
                           {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c', 'key3': 'val1f', 'key4': 'val2f'},
                           {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c', 'key3': 'val1g', 'key4': 'val2g'}]
        assert resp == expected_result
    
    def test_outer_join_left_preserve(self):
        """
        #### Function: 
            - Sftocsv.outer_join
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict]
            -@left_key: 'lkey' 
            -@right_key: 'rkey' 
            -@side: 'left'
            -@preserve_inner_key: True
        #### Expected Behaviour: 
            - The left_list is selected as the outer list, the right_list as the inner. 
            - the lists are looped through, the first record in the outer list matches the second in the inner, 
                these are combined and added the returned list. 
            - the second record in the outer list doesn't match anything, so it's just added the resulting list 
            - the third record in the outer matches 2 records in the inner, this results in 2 records added to the resulting list 
            - because preserve_inner_key is true, it keeps the inner key in each resulting record 
        #### Assertions: 
            - The returned list is as expected
        """
        input_left = [{'key1': 'val1a', 'key2': 'val2a', 'lkey': 'a'},
                      {'key1': 'val1b', 'key2': 'val2b', 'lkey': 'b'},
                      {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c'}]
        input_right = [{'key3': 'val1d', 'key4': 'val2d', 'rkey': 'd' },
                       {'key3': 'val1e', 'key4': 'val2e', 'rkey': 'a'},
                       {'key3': 'val1f', 'key4': 'val2f', 'rkey': 'c'},
                       {'key3': 'val1g', 'key4': 'val2g', 'rkey': 'c'}] 
        resp = Sftocsv.outer_join(left_list=input_left, right_list=input_right, 
                                  left_key='lkey', right_key='rkey', side='left', preserve_innner_key=True)
        expected_result = [{'key1': 'val1a', 'key2': 'val2a', 'lkey': 'a', 'key3': 'val1e', 'key4': 'val2e', 'rkey': 'a'},
                           {'key1': 'val1b', 'key2': 'val2b', 'lkey': 'b'},
                           {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c', 'key3': 'val1f', 'key4': 'val2f', 'rkey': 'c'},
                           {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c', 'key3': 'val1g', 'key4': 'val2g', 'rkey': 'c'}]
        assert resp == expected_result

    def test_outer_join_right(self):
        """
        #### Function: 
            - Sftocsv.outer_join
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict]
            -@left_key: 'lkey' 
            -@right_key: 'rkey' 
            -@side: 'right'
            -@preserve_inner_key: False
        #### Expected Behaviour: 
            - The right_list is selected as the outer list, the left_list as the inner. 
            - the lists are looped through, the first record in the outer list matches the second in the inner, 
                these are combined and added the returned list. 
            - the second record in the outer list doesn't match anything, so it's just added the resulting list 
            - the third record in the outer matches 2 records in the inner, this results in 2 records added to the resulting list 
            - because preserve_inner_key is False, the inner matching key is not kept 
        #### Assertions: 
            - The returned list is as expected
        """
        input_right = [{'key1': 'val1a', 'key2': 'val2a', 'rkey': 'a'},
                      {'key1': 'val1b', 'key2': 'val2b', 'rkey': 'b'},
                      {'key1': 'val1c', 'key2': 'val2c', 'rkey': 'c'}]
        input_left = [{'key3': 'val1d', 'key4': 'val2d', 'lkey': 'd' },
                       {'key3': 'val1e', 'key4': 'val2e', 'lkey': 'a'},
                       {'key3': 'val1f', 'key4': 'val2f', 'lkey': 'c'},
                       {'key3': 'val1g', 'key4': 'val2g', 'lkey': 'c'}] 
        resp = Sftocsv.outer_join(left_list=input_left, right_list=input_right, 
                                  left_key='lkey', right_key='rkey', side='right')
        expected_result = [{'key1': 'val1a', 'key2': 'val2a', 'rkey': 'a', 'key3': 'val1e', 'key4': 'val2e'},
                           {'key1': 'val1b', 'key2': 'val2b', 'rkey': 'b'},
                           {'key1': 'val1c', 'key2': 'val2c', 'rkey': 'c', 'key3': 'val1f', 'key4': 'val2f'},
                           {'key1': 'val1c', 'key2': 'val2c', 'rkey': 'c', 'key3': 'val1g', 'key4': 'val2g'}]
        assert resp == expected_result

    def test_outer_join_right_preserve(self):
        """
        #### Function: 
            - Sftocsv.outer_join
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict] 
            -@left_key: 'lkey' 
            -@right_key: 'rkey' 
            -@side: 'left'
            -@preserve_inner_key: True
        #### Expected Behaviour: 
            - The right_list is selected as the outer list, the left_list as the inner. 
            - the lists are looped through, the first record in the outer list matches the second in the inner, 
                these are combined and added the returned list. 
            - the second record in the outer list doesn't match anything, so it's just added the resulting list 
            - the third record in the outer matches 2 records in the inner, this results in 2 records added to the resulting list 
            - because preserve_inner_key is true, it keeps the inner key in each resulting record 
        #### Assertions: 
            - The returned list is as expected
        """
        input_right = [{'key1': 'val1a', 'key2': 'val2a', 'rkey': 'a'},
                      {'key1': 'val1b', 'key2': 'val2b', 'rkey': 'b'},
                      {'key1': 'val1c', 'key2': 'val2c', 'rkey': 'c'}]
        input_left = [{'key3': 'val1d', 'key4': 'val2d', 'lkey': 'd' },
                       {'key3': 'val1e', 'key4': 'val2e', 'lkey': 'a'},
                       {'key3': 'val1f', 'key4': 'val2f', 'lkey': 'c'},
                       {'key3': 'val1g', 'key4': 'val2g', 'lkey': 'c'}] 
        resp = Sftocsv.outer_join(left_list=input_left, right_list=input_right, 
                                  left_key='lkey', right_key='rkey', side='right', preserve_innner_key=True)
        expected_result = [{'key1': 'val1a', 'key2': 'val2a', 'rkey': 'a', 'key3': 'val1e', 'key4': 'val2e', 'lkey': 'a'},
                           {'key1': 'val1b', 'key2': 'val2b', 'rkey': 'b'},
                           {'key1': 'val1c', 'key2': 'val2c', 'rkey': 'c', 'key3': 'val1f', 'key4': 'val2f', 'lkey': 'c'},
                           {'key1': 'val1c', 'key2': 'val2c', 'rkey': 'c', 'key3': 'val1g', 'key4': 'val2g', 'lkey': 'c'}]
        assert resp == expected_result

    def test_outer_join_full(self):
        """
        #### Function:
            - Sftocsv.outer_join
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict]
            -@left_key: 'lkey'
            -@right_key: 'rkey'
            -@side: 'full'
            -@preserve_inner_key: False
        #### Expected Behaviour: 
            - the 'outer' list is set to the left_list, it then goes through and matches against records in the right_list, 
                any match removes the right_list record index from the unmatched_inner set. 
            - After this matching process, the return_list should contain all left_list records and their matches against right_list,
                any unmatched records from the right_list then need to be added, and they are found using the indexes stored in the unmatched_inner
                set. They're appended the return_list and returned 
        #### Assertions: 
            - The returned lsit is as expected 
        """
        input_left = [{'key1': 'val1a', 'key2': 'val2a', 'lkey': 'a'},
                      {'key1': 'val1b', 'key2': 'val2b', 'lkey': 'b'},
                      {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c'}]
        input_right = [{'key3': 'val1d', 'key4': 'val2d', 'rkey': 'd' },
                       {'key3': 'val1e', 'key4': 'val2e', 'rkey': 'a'},
                       {'key3': 'val1f', 'key4': 'val2f', 'rkey': 'c'},
                       {'key3': 'val1g', 'key4': 'val2g', 'rkey': 'c'},
                       {'key3': 'val1h', 'key4': 'val2h', 'rkey': 'z'}] 
        resp = Sftocsv.outer_join(left_list=input_left, right_list=input_right, left_key='lkey', right_key='rkey', side='full')
        expected_result = [{'key1': 'val1a', 'key2': 'val2a', 'lkey': 'a', 'key3': 'val1e', 'key4': 'val2e'},
                           {'key1': 'val1b', 'key2': 'val2b', 'lkey': 'b'},
                           {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c', 'key3': 'val1f', 'key4': 'val2f'},
                           {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c', 'key3': 'val1g', 'key4': 'val2g'},
                           {'key3': 'val1d', 'key4': 'val2d', 'rkey': 'd' },
                           {'key3': 'val1h', 'key4': 'val2h', 'rkey': 'z'}]
        assert resp == expected_result

    def test_outer_join_full_preserve(self):
        """
        #### Function:
            - Sftocsv.outer_join
        #### Inputs: 
            -@left_list: list[dict]
            -@right_list: list[dict]
            -@left_key: 'lkey'
            -@right_key: 'rkey'
            -@side: 'full'
            -@preserve_inner_key: True
        #### Expected Behaviour: 
            - the 'outer' list is set to the left_list, it then goes through and matches against records in the right_list, 
                any match removes the right_list record index from the unmatched_inner set. 
            - After this matching process, the return_list should contain all left_list records and their matches against right_list,
                any unmatched records from the right_list then need to be added, and they are found using the indexes stored in the unmatched_inner
                set. They're appended the return_list and returned 
            - because the preserve_inner_key is true, it keeps the right_list key
        #### Assertions: 
            - The returned list is as expected 
        """
        input_left = [{'key1': 'val1a', 'key2': 'val2a', 'lkey': 'a'},
                      {'key1': 'val1b', 'key2': 'val2b', 'lkey': 'b'},
                      {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c'}]
        input_right = [{'key3': 'val1d', 'key4': 'val2d', 'rkey': 'd' },
                       {'key3': 'val1e', 'key4': 'val2e', 'rkey': 'a'},
                       {'key3': 'val1f', 'key4': 'val2f', 'rkey': 'c'},
                       {'key3': 'val1g', 'key4': 'val2g', 'rkey': 'c'},
                       {'key3': 'val1h', 'key4': 'val2h', 'rkey': 'z'}] 
        resp = Sftocsv.outer_join(left_list=input_left, right_list=input_right, left_key='lkey', right_key='rkey', side='full', preserve_innner_key=True)
        expected_result = [{'key1': 'val1a', 'key2': 'val2a', 'lkey': 'a', 'key3': 'val1e', 'key4': 'val2e', 'rkey': 'a'},
                           {'key1': 'val1b', 'key2': 'val2b', 'lkey': 'b'},
                           {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c', 'key3': 'val1f', 'key4': 'val2f', 'rkey': 'c'},
                           {'key1': 'val1c', 'key2': 'val2c', 'lkey': 'c', 'key3': 'val1g', 'key4': 'val2g', 'rkey': 'c'},
                           {'key3': 'val1d', 'key4': 'val2d', 'rkey': 'd' },
                           {'key3': 'val1h', 'key4': 'val2h', 'rkey': 'z'}]
        assert resp == expected_result

 